@@ -16,9 +16,14 @@ def add(
     key: "DataIndexKey",
     ignore: Optional["Ignore"] = None,
 ):
+    entry = build_entry(path, fs)
+    entry.key = key
+    index.add(entry)
+
     if not fs.isdir(path):
-        index[key] = build_entry(path, fs)
         return
 
-    for entry_key, entry in build_entries(path, fs, ignore=ignore):
-        index[(*key, *entry_key)] = entry
+    for entry in build_entries(path, fs, ignore=ignore):
+        assert entry.key is not None
+        entry.key = (*key, *entry.key)
+        index.add(entry)
