@@ -2,7 +2,7 @@ from dvc_data.hashfile.db import HashFile, HashFileDB
 
 
 def test_db(tmp_upath, as_filesystem):
-    odb = HashFileDB(as_filesystem(tmp_upath.fs), tmp_upath)
+    odb = HashFileDB(as_filesystem(tmp_upath.fs), str(tmp_upath))
 
     assert not odb.exists("123456")
     assert list(odb.all()) == []
